@@ -1,13 +1,13 @@
-use std::{sync::{Arc, Mutex}, collections::HashMap, fs::File, io::Read};
+use std::{sync::{Arc, Mutex}, collections::HashMap, fs::File, io::{Read, BufReader}};
 use bzip2::bufread::BzDecoder;
 use databuffer::DataBuffer;
-use crate::Cache;
+use crate::{Cache, CacheIndex};
 
 type ParserFun<T> = fn(DataBuffer) -> T;
 
 pub trait DefParser {
     fn parse_bytes(bytes: Vec<u8>) -> Self where Self: Sized {
-        return DefParser::parse_buff(DataBuffer::from_bytes(&bytes));
+        DefParser::parse_buff(DataBuffer::with_vec(bytes))
     }
 
     fn parse_buff(buffer: DataBuffer) -> Self;
@@ -91,16 +91,16 @@ pub trait DefParser {
   ```
  */
 pub struct DefProvider<T> {
-    pub cache: Arc<Mutex<Cache>>,
+    pub file_provider: FileProvider,
     pub index: u32,
     pub parser: Option<ParserFun<T>>,
     def_cache: HashMap<u32, T>
 }
 
 impl <T: DefParser> DefProvider<T> {
-    pub fn with(cache: Arc<Mutex<Cache>>, index: u32) -> Self {
+    pub fn with(cache: &Arc<Mutex<Cache>>, index: u32) -> Self {
         Self {
-            cache: cache.clone(),
+            file_provider: FileProvider::from(cache),
             index,
             parser: Some(T::parse_buff),
             def_cache: HashMap::new()
@@ -112,11 +112,10 @@ impl <T: DefParser> DefProvider<T> {
             return self.def_cache.get(&id).unwrap();
         }
 
-        let mut data_provider = FileProvider::from(self.cache.clone());
+        self.file_provider.index(self.index);
+        self.file_provider.archive(archive);
 
-        data_provider.index(self.index);
-        data_provider.archive(&archive.get_id());
-        let data = data_provider.request(&file.get_id());
+        let data = self.file_provider.request(file);
 
         let parse = self.parser.unwrap();
 
@@ -141,7 +140,7 @@ impl <T: DefParser> DefProvider<T> {
   use idx::Cache;
 
   let cache = Arc::from(Mutex::from(Cache::from_path("test_cache").unwrap()));
-  let mut data_provider = FileProvider::from(cache);
+  let mut data_provider = FileProvider::from(&cache);
   
   data_provider.index(19).archive(&6);
   let data = data_provider.request(&17); //Returns the raw data for file 17 in archive 6 of index 19.
@@ -153,25 +152,14 @@ pub struct FileProvider {
     cache: Arc<Mutex<Cache>>,
     index: u32,
     archive: u32,
-    data_file: Arc<Mutex<File>>,
+    data_file: Arc<Mutex<BufReader<File>>>,
     keys: Vec<i64>
 }
 
 impl FileProvider {
-    pub fn from(cache: Arc<Mutex<Cache>>) -> Self {
+    pub fn from(cache: &Arc<Mutex<Cache>>) -> Self {
         let dfile = match cache.lock() {
-            Ok(n) => match n.data_file.lock() {
-                Ok(n) => match n.try_clone() {
-                    Ok(n) => Arc::from(Mutex::from(n)),
-                    Err(e) => {
-                        panic!("Unable to obtain new file reference: {}", e);
-                    }
-                }
-
-                Err(e) => {
-                    panic!("Unable to lock data file reference: {}", e);
-                }
-            }
+            Ok(n) => n.data_file.clone(),
 
             Err(e) => {
                 panic!("Unable to lock cache: {}", e);
@@ -179,7 +167,7 @@ impl FileProvider {
         };
 
         Self {
-            cache,
+            cache: cache.clone(),
             index: 0,
             archive: 0,
             data_file: dfile,
@@ -193,7 +181,17 @@ impl FileProvider {
     }
 
     pub fn archive(&mut self, archive: &dyn ContainerIdProvider) -> &mut Self {
-        self.archive = archive.get_id();
+        if self.index == 0 {
+            self.archive = archive.get_id(None);
+            println!("WARNING: archive was set before the index was! IDX: {}, ARCHIVE: {}. This will break archive access via name hashes!", self.index, self.archive);
+            return self
+        }
+
+        {
+            let mut _cache = self.cache.lock().unwrap();
+            let index = _cache.index(self.index as usize).unwrap();
+            self.archive = archive.get_id(Some(index));
+        }
         self
     }
 
@@ -202,7 +200,7 @@ impl FileProvider {
     }
 
     pub fn request(&mut self, file: &dyn ContainerIdProvider) -> DataBuffer {
-        let file_id = file.get_id();
+        let file_id = file.get_id(None);
 
         let file_data = match self.cache.lock() {
             Ok(mut n) => match n.index(self.index as usize) {
@@ -226,7 +224,7 @@ impl FileProvider {
         };
 
         if file_data.len() != 0 {
-            return file_data;
+            file_data
         } else {
             self.load_requested_container_files();
 
@@ -239,7 +237,7 @@ impl FileProvider {
                         }
                         None => {
                             println!("Invalid archive supplied?");
-                            return DataBuffer::new()
+                            DataBuffer::new()
                         }
                     },
                     None => {
@@ -281,14 +279,13 @@ impl FileProvider {
         };
 
         if file_info.len() == 1 {
-            match archive.file_containers.get_mut(&file_info[0]) {
-                Some(n) => n.data = container_data,
-                None => return
+            if let Some(file_container) = archive.file_containers.get_mut(&file_info[0]) {
+                file_container.data = container_data;
             }
         } else {
             let mut file_sizes = Vec::<i32>::new();
             for _ in 0..(num_loops as usize) {
-                let mut offset = 0 as i32;
+                let mut offset = 0_i32;
                 for file_index in 0..(file_info.len() as usize){
                     offset += buffer.read_i32();
                     if file_sizes.len() == file_index {
@@ -304,15 +301,15 @@ impl FileProvider {
             let mut offset = 0;
             for _ in 0..(num_loops as usize) {
                 let mut data_read = 0;
-                for file_index in 0..(file_info.len()) {
+                for file_index in &file_info {
                     data_read += buffer.read_i32();
 
-                    match archive.file_containers.get_mut(&file_info[file_index]) {
+                    match archive.file_containers.get_mut(file_index) {
                         Some(n) => {
                             n.data.append(&mut container_data[(offset as usize)..((offset + data_read) as usize)].to_vec())
                         },
                         None => {
-                            println!("Unknown file id: {}", file_info[file_index]);
+                            println!("Unknown file id: {}", file_index);
                             continue;
                         }
                     }
@@ -360,7 +357,7 @@ impl FileProvider {
         };
 
         for file in container.file_indices.iter() {
-            file_info.push(file.clone());
+            file_info.push(*file);
         }
 
         file_info
@@ -368,18 +365,24 @@ impl FileProvider {
 }
 
 pub trait ContainerIdProvider {
-    fn get_id(&self) -> u32;
+    fn get_id(&self, _: Option<&mut CacheIndex>) -> u32;
 }
 
-impl ContainerIdProvider for str {
-    fn get_id(&self) -> u32 {
-        return get_name_hash(&self);
+impl ContainerIdProvider for String {
+    fn get_id(&self, idx: Option<&mut CacheIndex>) -> u32 {
+        let hash = get_name_hash(&self);
+
+        if let Some(index) = idx {
+            index.get_container_by_name_hash(hash)
+        } else {
+            hash
+        }
     }
 }
 
 impl ContainerIdProvider for u32 {
-    fn get_id(&self) -> u32 {
-        return self.clone();
+    fn get_id(&self, _: Option<&mut CacheIndex>) -> u32 {
+        *self
     }
 }
 
@@ -396,10 +399,10 @@ fn get_name_hash(name: &str) -> u32 {
 }
 
 pub(crate) fn decompress_container_data(packed_data: Vec<u8>) -> Option<Vec<u8>> {
-    let mut data = DataBuffer::from_bytes(&packed_data);
+    let mut data = DataBuffer::with_vec(packed_data);
     let mut unpacked = Vec::<u8>::new();
 
-    if packed_data.is_empty() {
+    if data.len() == 0 {
         return Some(Vec::new());
     }
 
@@ -408,33 +411,29 @@ pub(crate) fn decompress_container_data(packed_data: Vec<u8>) -> Option<Vec<u8>>
 
     if container_size > 5000000 {
         println!("Invalid container size! {}", container_size);
-        return None;
+        None
     } else {
         match compression {
             0 => { //Uncompressed
-                for _ in 0..container_size {
-                    unpacked.push(data.read_u8());
-                }
-        
-                return Some(unpacked);
+                let trim_at = data.get_rpos();
+                let mut raw = data.deconstruct();
+
+                raw.drain(..trim_at);
+                Some(raw)
             },
 
             1 => { //Bzip2 (supposedly)
                 let decompressed_size = data.read_u32();
-                let mut current_index: usize = 0;
                 let trim_at = data.get_rpos() - 4;
 
-                let mut trimmed_data = data.to_bytes();
-                trimmed_data.retain(|_| {
-                    current_index +=1;
-                    current_index - 1 >= trim_at
-                });
+                let mut trimmed_data = data.deconstruct();
+                trimmed_data.drain(..trim_at);
 
                 //Re-add header jagex strips.
-                trimmed_data[0] = 'B' as u8;
-                trimmed_data[1] = 'Z' as u8;
-                trimmed_data[2] = 'h' as u8;
-                trimmed_data[3] = '1' as u8;
+                trimmed_data[0] = b'B';
+                trimmed_data[1] = b'Z';
+                trimmed_data[2] = b'h';
+                trimmed_data[3] = b'1';
 
                 match BzDecoder::new(&trimmed_data[..]).read_to_end(&mut unpacked) {
                     Ok(_) => {},
@@ -443,21 +442,17 @@ pub(crate) fn decompress_container_data(packed_data: Vec<u8>) -> Option<Vec<u8>>
                     }
                 }
 
-                assert_eq!(decompressed_size, unpacked.len() as u32, "packed size: {}, decompressed correct: {}, current decompressed: {}", packed_data.len(), decompressed_size, unpacked.len());
-                return Some(unpacked);
+                assert_eq!(decompressed_size, unpacked.len() as u32);
+                Some(unpacked)
             },
 
             _ => { //DEFLATE/Gzip/Zip
                 let decompressed_size = data.read_u32();
-                let mut current_index: usize = 0;
                 data.set_rpos(data.get_rpos() + 10);
                 let trim_at = data.get_rpos();
 
-                let mut trimmed_data = data.to_bytes();
-                trimmed_data.retain(|_| {
-                    current_index +=1;
-                    current_index - 1 >= trim_at
-                });
+                let mut trimmed_data = data.deconstruct();
+                trimmed_data.drain(..trim_at);
 
                 unpacked = match inflate::inflate_bytes(&trimmed_data) {
                     Ok(n) => n,
@@ -467,8 +462,8 @@ pub(crate) fn decompress_container_data(packed_data: Vec<u8>) -> Option<Vec<u8>>
                     }
                 };
 
-                assert_eq!(decompressed_size, unpacked.len() as u32, "packed size: {}, trimmed size: {}, decompressed correct: {}, current decompressed: {}", packed_data.len(), trimmed_data.len(), decompressed_size, unpacked.len());
-                return Some(unpacked);
+                assert_eq!(decompressed_size, unpacked.len() as u32);
+                Some(unpacked)
             }
         }
     }
